/*
  fragrep c++ rebuild */

// std and iostream headers
#include <ctype.h>
#include <iostream>
#include <fstream>
#include <list>
#include <stdio.h>
#include <map>
#include <string>

// libsequence headers for reading fasta files
//#include <Sequence/Fasta.hpp>
//#include <Sequence/SeqRegexes.hpp>
//#include <Sequence/SeqProperties.hpp>

// fragrep's own includes
#include "pattern_matching.h"

#include "fasta.h"

#define CHUNK_SIZE 4096


/** \mainpage

welcome to fragrep.

*/


using namespace fragrep;	


std::ifstream ifs;
std::list<std::pair<char*,int> > chunk_list;
int last_chunk_length = 0;
int total_length = 0;
char* chunk_block;
bool final_chunk;
char fasta_header[CHUNK_SIZE];

/// \c genome sequence
char* T;

std::string T_name;

/// \c k sequence fragments
char** s;


/// n[j] is the length of s[j]
int* n;

/// N is the length of T
int N;

/// m[j] number of mismatches allowed for occurences of s[j]
int* m;

/// u[j] is an upper bound for the gap length between s[j-1] and s[j].
int* u;

/// l[j] is a lower bound for the gap length between s[j-1] and s[j].
int* l;

/// number of binding site fragments
int k;

void fill_reverse_segments();
void print_help_message();

/// Are the parameters passed to the program legal?
bool params_good;

/// number of command-line options
int num_of_opts;

/// consider reverse complemented segments also?
bool reverse;

/// print simplifed output?
bool simple_output;

/// false if -u option (unweighted) is given.
bool weighted;

/// specify extra output to stderr
bool verbose;

/// num of characters of extra output before a matching block
int pref_length;

/// num of characters of extra output after a matching block
int suff_length;

/// read over all useless characters such as spaces, tabs and line
/// breaks
void skip(std::ifstream& ifs)
{
	if (!ifs.good())
		return;
	char ch = 0;
	ifs.get(ch);
	while (ifs.good() && ch<=32)
		ifs.get(ch);
	if (ifs.good())
		ifs.unget();
}

bool is_int(char* s)
{
	int i=0;
	while (s[i]!=0)
	{
		if (s[i]<'0' || s[i]>'9')
			return false;
		i++;
	}
	return true;
}

bool is_flt(char* s)
{
	int i=0;
	int dot_cnt = 0;
	while (s[i]!=0)
	{
		if (s[i]=='.')
			dot_cnt++;
		if (dot_cnt>1)
		{
			return false;
		}
		if ((s[i]<'0' || s[i]>'9') && s[i]!='.')
		{
			return false;
		}
		i++;
	}
	return true;
}




bool read_parameters(int argc, char** argv)
{
	if (argc<3)
	{
		print_help_message();		
		return false;
	}
	
	weighted = false;
	params_good = true;
	reverse = false;
	simple_output = false;
	num_of_opts = 0;
	int i;
	for (i=1; i<argc && argv[i][0]=='-'; i++)
	{
		num_of_opts++;
		switch (argv[i][1])
		{
			case 's' : {
				if (strlen(argv[i])>2)
				{
					std::cerr<<"Unknown option: "<<argv[i]<<"\n";
					return false;
				}
				if (argc<i+3)
				{
					std::cerr<<"Option -s takes two integer parameters.\n";
					return false;
				}
				if (!is_int(argv[i+1]) || !is_int(argv[i+2]))
				{
					std::cerr<<"Option -s takes two integer parameters.\n";
					return false;
				}
				pref_length = atoi(argv[i+1]);
				suff_length = atoi(argv[i+2]);				
				num_of_opts += 2;
				break;
			}
			case 'u' : {
				weighted = false;
				if (strlen(argv[i])>2)
				{
					std::cerr<<"Unknown option: "<<argv[i]<<"\n";
					return false;
				}
				break;
			}
			case 'v' : {
				verbose = true;
				if (strlen(argv[i])>2)
				{
					std::cerr<<"Unknown option: "<<argv[i]<<"\n";
					return false;
				}
				break;
			}
			case 'r' : {
				reverse = true;
				if (strlen(argv[i])>2)
				{
					std::cerr<<"Unknown option: "<<argv[i]<<"\n";
					return false;
				}
				break;
			} case 'S' : {
				simple_output = true;
				if (strlen(argv[i])>2)
				{
					std::cerr<<"Unknown option: "<<argv[i]<<"\n";
					return false;
				}
				break;
			} case 'H' : {
				print_help_message();
				break;
			} case 'w' : {
				weighted = true;
				if (strlen(argv[i])>3)
				{
					std::cerr<<"Unknown option: "<<argv[i]<<"\n";
					return false;
				}
				if (strlen(argv[i])>2)
					switch (argv[i][2])
					{
						case 'm' :
							break;
						case 'p':
							break;
						default:
							std::cerr<<"ignoring unknown weight type "<<argv[i][2]<<std::endl;
							break;
					}
				
				break;
			}
			default: {
				std::cerr<<"unknown option: "<<argv[i][1]<<std::endl;
				return false;
			}
		}
	}
	return params_good;
}

void print_help_message()
{
	std::cerr<<"Use fragrep as follows:\n\n";
	std::cerr<<"fragrep [options] <motif-file> [<genome-file>],\n\n";
	std::cerr<<"where <motif-file> is a filename as well as <genome-file>.\n";
	std::cerr<<"if <genome-file> is not specified, stdin is read instead, which is expected to be in fasta format.\n";
	std::cerr<<"valid options are:\n";
	std::cerr<<"-u (unweighted) to switch off weighting and optimize for intersection cardinality;\n";
	std::cerr<<"-w select weighting based on based on p-values\n";
	std::cerr<<"-r include reverse complemented fragments so that both 3' and 5' occurences may be part of a cluster\n";
	std::cerr<<"-s <prefixlength> <suffixlength> to extend extraction of result sequences by a specified number of nucleotides\n";
	std::cerr<<"-S use simplified output (leave out aligned query sequences and print reverse sequences for reverse matches)\n";
	std::cerr<<"   output produced with -S can be plugged into alignment programs such as clustalw immediately, but is less legible.\n";
	std::cerr<<"\n";
}

bool read_data(int argc, char** argv)
{
	
	int j;
	params_good = true;

	if (argc-num_of_opts<2)
	{
		print_help_message();
		params_good = false;
	}
	
	// 0. READ NUMBER OF FRAGMENTS
	std::ifstream frag_file(argv[num_of_opts+1]);
	if (!frag_file.good())
	{
		std::cerr<<"could not open file "<<argv[num_of_opts+1]<<"\n";
		return false;
	}
	skip(frag_file);	
	char read_buf[8192];
	// read fragment name
	read_buf[0]=0;
	read_buf[8191] = 0;
	frag_file.getline(read_buf,8191);
	if (!is_int(read_buf))
	{
		std::cerr<<"expected an integer in pattern file.\n";
		exit(-1);
	}
	k = atoi(read_buf);
	s = new char*[k];
	m = new int[k];
	n = new int[k];
	u = new int[k];
	l = new int[k];
	
	skip(frag_file);		
		
	// I. READ FRAGMENT SEQUENCES s_1,...,s_k from the fragment-file

	//while (frag_file.good())
	for (j=0; j<k && frag_file.good(); j++)
	{
		// read lower length bound
		read_buf[0]=0;
		read_buf[8191] = 0;
		frag_file.getline(read_buf,8192,' ');
		if (!is_int(read_buf))
		{
			std::cerr<<"expected an integer in pattern file.\n";
			exit(-1);
		}
		l[j] = atoi(read_buf);
		skip(frag_file);

		// read upper length bound
		read_buf[0]=0;
		read_buf[8191] = 0;
		frag_file.getline(read_buf,8192,' ');
		if (!is_int(read_buf))
		{
			std::cerr<<"expected an integer in pattern file.\n";
			exit(-1);
		}
		u[j] = atoi(read_buf);
		skip(frag_file);
		
		// read fragment sequence, currently delimited to 8192 characters
		read_buf[0]=0;
		read_buf[8191] = 0;
		int s_len;
		frag_file.getline(read_buf,8191,' ');
		s_len = strlen(read_buf);
		s[j] = new char[s_len+1];
		strcpy(s[j],read_buf);
		n[j] = strlen(s[j]);
		skip(frag_file);
		
		// read maximum edit distance for the fragment
		read_buf[0]=0;
		read_buf[8191] = 0;
		frag_file.getline(read_buf,8191);
		if (!is_int(read_buf))
		{
			std::cerr<<"expected an integer in pattern file.\n";
			exit(-1);
		}
		m[j] = atoi(read_buf);
		skip(frag_file);
	}

	// II. INIT GENOME SEQUENCES T
	//if (!init_read_genome(argv[num_of_opts+2]))
	//	return false;


	//if (reverse)
	//  revert_segments();

	return true;
	
}

void revert_segments()
{

	if (!reverse)
		return;
	
	int i;
	
	char* compl_table = new char[256];

	for(i=0; i<256; i++) compl_table[i]=(int)'-';
	
	compl_table[(int)'A']='T';
	compl_table[(int)'C']='G';
	compl_table[(int)'G']='C';
	compl_table[(int)'T']='A';
	compl_table[(int)'R']='Y';
	compl_table[(int)'Y']='R';
	compl_table[(int)'M']='K';
	compl_table[(int)'K']='M';
	compl_table[(int)'S']='S';
	compl_table[(int)'W']='W';
	compl_table[(int)'H']='D';
	compl_table[(int)'B']='V';
	compl_table[(int)'V']='B';
	compl_table[(int)'D']='H';
	compl_table[(int)'N']='N';
	compl_table[(int)'a']='t';
	compl_table[(int)'c']='g';
	compl_table[(int)'g']='c';
	compl_table[(int)'t']='a';
	compl_table[(int)'r']='y';
	compl_table[(int)'y']='r';
	compl_table[(int)'m']='k';
	compl_table[(int)'k']='m';
	compl_table[(int)'s']='s';
	compl_table[(int)'w']='w';
	compl_table[(int)'h']='d';
	compl_table[(int)'b']='v';
	compl_table[(int)'v']='b';
	compl_table[(int)'d']='h';
	compl_table[(int)'n']='n';
	
	int j, iota;
	// k is divisible by two, since it has been mult by two due to
	// reverse option before.
// 	int k_2 = k/2;
// 	for (j=0; j<k_2; j++)
// 	{
// 		int n_j = n[j];
// 		s[k_2+j] = new char[n[j]+1];
// 		n[k_2+j] = n[j];
// 		m[k_2+j] = m[j];
// 		int len = strlen(s_names[j])+5;
// 		s_names[k_2+j] = new char[len];
// 		strcpy(s_names[k_2+j],s_names[j]);
// 		strcpy(s[k_2+j],s[j]);
// 		strcat(s_names[k_2+j],"-rev");
// 		for (iota=0; iota<n_j; iota++)
// 			s[k_2+j][iota] = compl_table[s[j][n_j-iota-1]];
// 		s[k_2+j][n_j] = (char)0;
// 	}
	//for (j=0; j<k; j++)
	//	std::cerr<<s_names[j]<<" "<<s[j]<<std::endl;
}


void free_all()
{
	/*int i,j;
	//delete[] T;
	T = NULL;;
	for (j=0; j<k; j++)
		delete[] s[j];
	delete[] s;
	s = NULL;
	delete[] n;
	n = NULL;
	delete[] l;
	l = NULL;
	delete[] u;
	u = NULL;
	delete[] m;
	m = NULL;
	*/
}

int main(int argc, char** argv)
{
	
	T = NULL;;
	s = NULL;
	n = NULL;
	l = NULL;
	u = NULL;
	m = NULL;

	if (!read_parameters(argc,argv))
	{
		exit(0);
		return 0;
	}		
	if (!params_good)
	{
		print_help_message();
		exit(0);
		return 0;
	}
	if (!read_data(argc,argv))
	{
		exit(0);
		return 0;
	}

	int jj;

	pattern_matching pm(k, s, n, l, u, m);
	pm.set_output_region(pref_length,suff_length);
	pm.set_simple_output(simple_output);
	pm.set_verbose(verbose);
	// by default, fastafile reads from stdin.
	fasta_reader fastafile;
	if (argc-num_of_opts>2)
		fastafile.open(argv[num_of_opts+2]);

	std::string header,seq;
	if (verbose)
		std::cerr<<"checking for next sequence...\n";

	while (fastafile.get_seq (header, seq)) {
		std::string T_name = header;
		//char* T_ = new char[seq.length()+1];
		//strcpy(T_,seq.c_str());
		if (verbose)
			std::cerr<<"searching sequence "<<header<<"\n";
		pm.set_genome(seq.c_str(),header,seq.length());
		pm.run();
		if (reverse)
		{
			if (verbose)
				std::cerr<<"searching reverse complement...\n";
			pm.reverse_compl();
			pm.run();
			// in case this was not the last genome sequence, we nedd to
			// restore the original status...
			pm.reverse_compl();
		}
		if (verbose)
			std::cerr<<"checking for next sequence...\n";
	} 
	if (verbose)
		std::cerr<<"done."<<header<<"\n";

	pm.clear();
	free_all();
	
	return 0;
	
}



