// bbq - Copyright (C) 2005, Axel Mosig, University of Leipzig. See
// main.cpp for details.

#include <limits.h>
#include <float.h>
#include <stdlib.h>

#include <iostream>
#include <fstream>
#include <ios>
#include <math.h>

#include <list>
#include <vector>
#include <string>
#include <algorithm>
#include "triple.h"

#include "pattern_matching.h"


namespace fragrep {
	
/// A cluster triple is a triplet (p,e,j), where p is the position, e
/// the edit distance and j the fragment index of an occurence in some
/// T_i.
typedef std::triple< int,int,int > cluster_triple;
typedef std::set<cluster_triple> cluster_set;
typedef std::multimap<int,cluster_triple> cluster_map;


#define max(x,y) ( ( (x) > (y) ) ? (x) : (y) )

typedef std::list< std::pair<int,float> > open_interval_list;

// ==============================================================
pattern_matching::pattern_matching()
// ==============================================================
{

	init();
	
	reverted  = false;
	simple_output = false;
	
	k = 0;
	s = NULL;
	n = NULL;
	m = NULL;	
	l = NULL;
	u = NULL;
	T = NULL;
	T_name = "";
}


// ==============================================================
pattern_matching::pattern_matching(char* T_, const std::string& T_name_, int N_, int k_, char** s_, int* n_, int* l_, int* u_, int* m_)
// ==============================================================
{

	init();
	
	reverted  = false;
	simple_output = false;
	k = k_;
	N = N_;
	s = s_;
	T = new char[N_+1000];
	strcpy(T,T_);
	T_name = T_name_;
	n = n_;
	m = m_;	
	l = l_;
	u = u_;

	match_reverse = false;
	reverted = false;
	prefix_length = 0;
	suffix_length = 0;

	compute_index_table();
	compute_match_table();
}


// ==============================================================
pattern_matching::pattern_matching(int k_, char** s_, int* n_, int* l_, int* u_, int* m_)
// ==============================================================
{

	init();
	
	reverted  = false;
	simple_output = false;
	k = k_;
	s = s_;
	n = n_;
	m = m_;	
	l = l_;
	u = u_;
	T = NULL;
	T_name = "";

	match_reverse = false;
	reverted = false;
	prefix_length = 0;
	suffix_length = 0;

	compute_index_table();
	compute_match_table();
}

// ==============================================================
void pattern_matching::init()
// ==============================================================
{
	compl_table = new char[256];

	int i;
	for(i=0; i<256; i++) compl_table[i]=(int)'*';
	
	compl_table[(int)'A']='T';
	compl_table[(int)'C']='G';
	compl_table[(int)'G']='C';
	compl_table[(int)'T']='A';
	compl_table[(int)'R']='Y';
	compl_table[(int)'Y']='R';
	compl_table[(int)'M']='K';
	compl_table[(int)'K']='M';
	compl_table[(int)'S']='S';
	compl_table[(int)'W']='W';
	compl_table[(int)'H']='D';
	compl_table[(int)'B']='V';
	compl_table[(int)'V']='B';
	compl_table[(int)'D']='H';
	compl_table[(int)'N']='N';
	compl_table[(int)'a']='t';
	compl_table[(int)'c']='g';
	compl_table[(int)'g']='c';
	compl_table[(int)'t']='a';
	compl_table[(int)'r']='y';
	compl_table[(int)'y']='r';
	compl_table[(int)'m']='k';
	compl_table[(int)'k']='m';
	compl_table[(int)'s']='s';
	compl_table[(int)'w']='w';
	compl_table[(int)'h']='d';
	compl_table[(int)'b']='v';
	compl_table[(int)'v']='b';
	compl_table[(int)'d']='h';
	compl_table[(int)'n']='n';
}
	
// ==============================================================
void pattern_matching::set_output_region(int pref_len, int suff_len)
// ==============================================================
{
	prefix_length = pref_len;
	suffix_length = suff_len;
}

// ==============================================================
void pattern_matching::set_reverse(bool rev)
// ==============================================================
{
	match_reverse = rev;
}
	
// ==============================================================
void pattern_matching::set_verbose(bool verb)
// ==============================================================
{
	verbose = verb;
}
	
// ==============================================================
void pattern_matching::set_simple_output(bool so)
// ==============================================================
{
	simple_output = so;
}
	

// ==============================================================
void pattern_matching::set_genome(const char* T_, const std::string& T_name_, int N_)
// ==============================================================
{
	if (T!=NULL)
	{
		delete[] T;
		T = NULL;
	}
	T = new char[N_+100];
	strcpy(T,T_);

	T_name = T_name_;
	N = N_;
}


// ==============================================================
int pattern_matching::get_next_occurence(int j, int up_to)
// ==============================================================
{

	if (up_to<done[j])
		return -1;
	if (up_to>N-n[j])
		up_to = N-n[j];
	
	int jj, m_cost;
	int m_j = m[j];
	int n_j = n[j];
	char* s_j = s[j];
	int ii = done[j];
	
	for (; ii<up_to; ii++)
	{
		m_cost = 0;
		for (jj=0; (jj<n_j) && (m_cost<=m_j); jj++)
		{
			//std::cout<<"("<<s_j[jj]<<T[ii+jj]<<")";
			if (match_table[index_table[(int)s_j[jj]]][index_table[T[ii+jj]]]==0)
				m_cost++;
		}
		if (m_cost<=m_j) {
			//std::cout<<"*\n";
			done[j] = ii+1;
			return ii;
		}
		//std::cout<<"\n";
	}
	done[j] = up_to+1;
	return -1;
	
}


// ==============================================================
bool pattern_matching::check_match_lists()
// ==============================================================
{

	int j;
	bool hope_left;

	// if there is no hope left that we can find any match for the
	// current occurence of s[j_inf], hope_left is set to false, which
	// gives a sppedup because the for loop in this procedure can be
	// terminated early. Initially, we have not lost all hopes for
	// success, so hope_left is set to true :)
	hope_left = true;
	
	match_list* match_list_j = &(match_lists[0]);
	match_list::iterator m_it = match_lists[0].begin();
	match_list::iterator m_end = match_lists[0].end();
	while (m_it != m_end)
	{
		m_it->second = true;
		m_it++;
	}
	for (j=1; j<k && hope_left; j++)
	{
		match_list_j = &(match_lists[j]);
		match_list* match_list_prev = &(match_lists[j-1]);

		m_it = match_list_j->begin();
		m_end = match_list_j->end();
		hope_left = false;
		while (m_it != m_end)
		{
			m_it->second = false;
			match_list::iterator p_it = match_list_prev->begin();
			match_list::iterator p_end = match_list_prev->end();
			while (p_it != p_end)
			{
				if (p_it->second)
				{
					int m_pos = m_it->first;
					int p_pos = p_it->first;
					int dist = m_pos-p_pos-n[j-1];
					if (dist>=l[j] && dist<=u[j])
					{
						m_it->second = true;
						hope_left = true;
					}	
				}
				p_it++;
			}
			m_it++;
		}
	}
	return hope_left;
	
}

// ==============================================================
bool pattern_matching::update_match_lists(int pos)
// ==============================================================
{
	// if there is no hope left that we can find any match for the
	// current occurence of s[j_inf], hope_left is set to false, which
	// gives a sppedup because the for loop in this procedure can be
	// terminated early. Initially, we have not lost all hopes for
	// success, so hope_left is set to true :)
	bool hope_left = true;
	int j;
	for (j=0; j<k && hope_left; j++)
	{
		// remove old matches from match_lists[j]
		int curr_l_bound_j = pos+l_bound[j];
		match_list* match_list_j = &(match_lists[j]);
		if (j!=j_inf)
		{
			match_list::iterator m_it = match_list_j->begin();
			match_list::iterator m_end = match_list_j->end();
			bool abort = false;
			if (match_list_j->empty())
				abort = true; // GODFUCKING GCC
			if (abort==false)
			{
				while ((m_it!=m_end) && !abort)
				{
					if (m_it->first < curr_l_bound_j)
					{
						m_it++;
						match_list_j->pop_front();
					} else {
						abort = true;
					}
				}
			}
			while (m_it!=m_end)
			{
				m_it->second = false;
				m_it++;
			}
			
			// add new matches to match_lists[j]
			int curr_u_bound_j = pos+u_bound[j];
			int j_pos;
			while ( (j_pos=get_next_occurence(j,curr_u_bound_j)) >= 0 )
			{
				match_list_j->push_back(match_pair(j_pos,false));
			}
		} else {
			match_list_j->clear();
			match_list_j->push_back(match_pair(pos,false));
		}
		hope_left = !match_list_j->empty();
	}
	return hope_left;
}

// ==============================================================
void pattern_matching::update_bounds(int pos)
// ==============================================================
{
	int j;
	int* l_bound_j = &(l_bound[0]);
	for (j=0; j<k; j++)
	{
		int new_b = pos + (*l_bound_j);
		if ( (new_b>=0) && (new_b>done[j]) )
			done[j] = new_b;
		l_bound_j++;
	}
}

// ==============================================================
void pattern_matching::print_result(result_list& result)
// ==============================================================
{

	number_of_matches++;
	
	/// PRINT HEADER
	int jj = 0;
	char dir_str[6];
	double E_val=0., p_val=1.;
	result_list::iterator r_it = result.begin();
	result_list::iterator r_end = result.end();
	for (jj=0 ; r_it!=r_end ; ++r_it, jj++ )
	{
		p_val *= ((double)(u[jj]-l[jj]+1)/(double)N) * r_it->p_val;			
	}
	E_val = -log(p_val);
	if (!simple_output)
	{
		if (reverted)
			strcpy(dir_str,"-bwd");
		else
			strcpy(dir_str,"-fwd");
	} else {
		if (reverted)
			strcpy(dir_str,"-");
		else
			strcpy(dir_str,"+");
	}
	if (!simple_output)
		std::cout<<T_name<<dir_str<<":pos"<<(result.begin()->pos+1);
	else
		std::cout<<T_name<<dir_str<<(result.begin()->pos+1);		
	if (!simple_output)
		std::cout<<" weight="<<E_val<<" p-value="<<p_val;
	std::cout<<"\n";

	/// PRINT SEQUENCE
	int seq_beg = result.front().pos-prefix_length;
	if (seq_beg<0)
		seq_beg = 0;
	int seq_end = result.back().pos+n[k-1]+suffix_length;
	if (seq_end > N)
		seq_end = N;
	int i;
	std::cout<<" ";
	if (reverted && simple_output) {
		for (i=seq_end-1; i>=seq_beg; i--)
			std::cout<<compl_table[T[i]];
	} else {
		for (i=seq_beg; i<seq_end; i++)
			std::cout<<T[i];
	}
	std::cout<<"\n";

	/// PRINT FRAGMENTS BELOW SEQUENCE
	if (!simple_output)
	{
		std::cout<<">matchsequence\n";
		int suff_len = suffix_length;
		if (result.begin()->pos<suff_len)
			suff_len = result.begin()->pos;
		std::cout<<" ";
		for (i=0; i<suff_len; i++)
			std::cout<<"-";
		std::cout<<s[0];
		r_it = result.begin();
		int prev_pos = r_it->pos;
		r_it++;
		r_end = result.end();
		for (jj=1 ; r_it!=r_end ; ++r_it, jj++ )
		{
			int gap_len = r_it->pos-prev_pos-n[jj-1];
			prev_pos = r_it->pos;
			for (i=0; i<gap_len; i++)
				std::cout<<"-";
			std::cout<<s[jj];
		}
		int suff_beg = result.back().pos + n[k-1];
		int suff_end = result.back().pos + n[k-1] + suffix_length;
		if (suff_end >= N)
			suff_end = N;
		for (jj=suff_beg; jj<suff_end; jj++)
			std::cout<<"-";
		std::cout<<"\n";
	}
}

// ==============================================================
void pattern_matching::output_matches(int j, result_list& result)
// ==============================================================
{
	if (j>=0) {
		match_list::iterator m_it = match_lists[j].begin();
		match_list::iterator m_end = match_lists[j].end();
		for (; m_it!=m_end; ++m_it)
		{
			if (m_it->second)
			{
				int pos = m_it->first;
				int diff, n_pos;
				bool valid = true;
				if (!result.empty())
				{
					n_pos = result.front().pos;
					diff = n_pos-pos-n[j];
					if (diff<l[j+1] || diff>u[j+1])
						valid = false;
				}
				if (valid)
				{
					result.push_front(result_item(pos,p_value(j,pos),E_value(j,pos)));
					output_matches(j-1,result);
					result.pop_front();
				}
			}
		}
	} else {
		print_result(result);

	}		
}

// ==============================================================
void pattern_matching::output_matches()
// ==============================================================
{
	result_list result;
	result.clear();
	output_matches(k-1,result);
}

// ==============================================================
void pattern_matching::match()
// ==============================================================
{

	int j_inf_pos;
	
	while ( (j_inf_pos=get_next_occurence(j_inf,N)) >= 0 )
	{
		update_bounds(j_inf_pos);
		if (update_match_lists(j_inf_pos))
			if (check_match_lists())
				output_matches();
	}
	
}

// ==============================================================
void pattern_matching::run()
// ==============================================================
{
	
	int i, j, ii, jj, eta;

	number_of_matches = 0;
	
	get_freq();
	j_inf = get_j_inf();

	done = new int[k];
	l_bound = new int[k];
	u_bound = new int[k];
	match_lists = new match_list[k];
	match_lists[0].clear();
	
	l_bound[j_inf] = 0;
	u_bound[j_inf] = 0;
	
	for (j=0; j<k; j++)
	{
		done[j] = 0;
		match_lists[j] = match_list();
		match_lists[j].clear();
	}
	for (j=j_inf; j>0; j--)
	{
		l_bound[j-1] = l_bound[j]-u[j]-n[j-1];
		u_bound[j-1] = u_bound[j]-l[j]-n[j-1];
	}
	for (j=j_inf+1; j<k; j++)
	{
		l_bound[j] = l_bound[j-1]+l[j]+n[j];
		u_bound[j] = u_bound[j-1]+u[j]+n[j];
	}
	
	match();

	if (verbose)
		std::cerr<<number_of_matches<<" match(es) found.\n";
	//delete[] done;
	//delete[] l_bound;
	//delete[] u_bound;
	//delete[] match_lists;
	
}


/// compute the reverse complement of sequence S
//==============================================================
	void pattern_matching::reverse_compl(char* S, int len_S)
// ==============================================================
{
	int iota;
	int half = len_S/2;
	for (iota=0; iota<half; iota++)
	{
		char tmp = compl_table[S[iota]];
		S[iota] = compl_table[S[len_S-iota-1]];
		S[len_S-iota-1] = tmp;
	}
	if (len_S%2==1)
	{
		S[half] = compl_table[S[half]] ;
	}
}
	
/// compute the reverse complement of the patterns, so that in effect
/// the genome is searched in backward direction for L-occurences.
//==============================================================
	void pattern_matching::reverse_compl()
// ==============================================================
{

	int i;
	int j, iota;

	reverted = !reverted;
	
	for (j=0; j<k; j++)
	{
		int n_j = n[j];
		int half = n_j/2;
		for (iota=0; iota<half; iota++)
		{
			char tmp = compl_table[s[j][iota]];
			s[j][iota] = compl_table[s[j][n_j-iota-1]];
			s[j][n_j-iota-1] = tmp;
		}
		if (n_j%2==1)
		{
			s[j][half] = compl_table[s[j][half]] ;
		}
	}
	int k_2 = k/2;
	for (j=0; j<k_2; j++)
	{
		char* s_tmp = s[j];
		s[j] = s[k-j-1];
		s[k-j-1] = s_tmp;
		
		int n_tmp = n[j];
		n[j] = n[k-j-1];
		n[k-j-1] = n_tmp;
		
		int m_tmp = m[j];
		m[j] = m[k-j-1];
		m[k-j-1] = m_tmp;		
	}

	k_2 = (k-1)/2+1;
	for (j=1; j<k_2; j++)
	{
		int u_tmp = u[j];
		u[j] = u[k-j];
		u[k-j] = u_tmp;
		
		int l_tmp = l[j];
		l[j] = l[k-j];
		l[k-j] = l_tmp;
	}

}


// ==============================================================
void pattern_matching::compute_match_table()
// ==============================================================
{

	int i, j, k, flag_match;

	bool case_sense = false;
	unsigned char asymmetric = 1;

	match_table = new unsigned char*[31];
	for(i=0; i<31; i++) 
		match_table[i] = new unsigned char[31];

	unsigned char letters[][8] = {
		/*	 A,C,G,T, a,c,g,t */
		
		{1,0,0,0, 0,0,0,0},    /* A */
	   {0,1,0,0, 0,0,0,0},    /* C */
	   {0,0,1,0, 0,0,0,0},    /* G */
	   {0,0,0,1, 0,0,0,0},    /* T */
	   {1,0,1,0, 0,0,0,0},	  /* R */
	   {0,1,0,1, 0,0,0,0},    /* Y */
	   {1,1,0,0, 0,0,0,0},    /* M */
	   {0,0,1,1, 0,0,0,0},    /* K */
	   {0,1,1,0, 0,0,0,0},    /* S */
	   {1,0,0,1, 0,0,0,0},    /* W */
	   {1,1,0,1, 0,0,0,0},    /* H */
	   {0,1,1,1, 0,0,0,0},    /* B */
	   {1,1,1,0, 0,0,0,0},    /* V */
	   {1,0,1,1, 0,0,0,0},    /* D */
	   {1,1,1,1, 0,0,0,0},    /* N */
		
		{0,0,0,0, 1,0,0,0},    /* a */
	   {0,0,0,0, 0,1,0,0},    /* c */
	   {0,0,0,0, 0,0,1,0},    /* g */
	   {0,0,0,0, 0,0,0,1},    /* t */
	   {0,0,0,0, 1,0,1,0},	  /* r */
	   {0,0,0,0, 0,1,0,1},    /* y */
	   {0,0,0,0, 1,1,0,0},    /* m */
	   {0,0,0,0, 0,0,1,1},    /* k */
	   {0,0,0,0, 0,1,1,0},    /* s */
	   {0,0,0,0, 1,0,0,1},    /* w */ 
	   {0,0,0,0, 1,1,0,1},    /* h */
	   {0,0,0,0, 0,1,1,1},    /* b */
	   {0,0,0,0, 1,1,1,0},    /* v */ 
	   {0,0,0,0, 1,0,1,1},    /* d */
	   {0,0,0,0, 1,1,1,1},    /* n */
 
		{0,0,0,0, 0,0,0,0}     /* all other chars; */
	};

	for(i=0; i<31; i++)
	{
		
		for(j=0; j<31; j++)
		{
			
			match_table[i][j] = 0;                       /* Default: no intersection present; */
			match_table[j][i] = 0;
			
			flag_match = 0;
			
			for(k=0; k<8; k++) {                              /* watch for intersections; */
				
				if(case_sense) {
					if(letters[i][k] + letters[j][k] == 2) {    /* mind capital letters; */
						flag_match = 1;
						break;
					}
				}
				
				else {                                         /* do not casre about capitals; */
					if(k<4) {
						if(letters[i][k] + letters[j][k] + letters[i][k+4] + letters[j][k+4] >= 2) {
							flag_match = 1;
							break;
						}
					}
					else {
						if(letters[i][k] + letters[j][k] + letters[i][k-4] + letters[j][k-4] >= 2) {
							flag_match = 1;
							break;
						}
					}
				}
			}
			
			if(flag_match==1) {
				match_table[i][j] = 1;                    /* there was a match -> non-empty intersection;    */
				match_table[j][i] = 1;
			}
		}
		
	}

	if (asymmetric)
	{
		for(i=0; i<31; i++)
		{
			match_table[i][14] = 0;
			match_table[i][29] = 0;
		}
		match_table[14][14] = 1;
		match_table[14][29] = 1;
		match_table[29][14] = 1;
		match_table[29][29] = 1;
	}
}


// ==============================================================
void pattern_matching::clear()
// ==============================================================
{
	int i,j,nu;
}


// ==============================================================
pattern_matching::~pattern_matching()
// ==============================================================
{
	clear();
}


// ==============================================================
int pattern_matching::get_j_inf()
// ==============================================================
{
	int j;
	int my_j_inf = 0;
	double E_inf = E_value(0,-1);
	
	for (j=1; j<k; j++)
	{
		double E_val = E_value(j,-1);
		if (E_val<E_inf)
		{
			my_j_inf = j;
			E_inf = E_val;
		}
	}

	return my_j_inf;
	
}


// ==============================================================
double pattern_matching::E_value(int j, int pos)
// ==============================================================
{
	double p_val = p_value(j,pos);

	/// non-positive p-values don't make sense -- thus we better rank those by an E-value of 0.
	if (p_val<=0.)
		return 0.;
	
	return -log(p_val);
}


// ==============================================================
double pattern_matching::p_value(int j, int pos)
// ==============================================================
{

	if (j>=k)
		return 1.;
	
	int jj, kk;
	double seq_N, L;
	seq_N = (double)N;	
	if (j==0)
		L = (double)N;
	else
		L = (double)(u[j]-l[j]);
	if (L<1.)
		L = 1.;

	double p = 1.;
	
	const char* txt;
	if (pos>=0)
		txt = T;
	else {
		txt = s[j];
		pos = 0;
	}
	//if (s[j][0]==txt[pos])
	if (match_table[index_table[(int)s[j][0]]][index_table[txt[pos]]]==1)	
		p *= (double)frequencies[index_table[s[j][0]]]/seq_N;
	int kk_end = n[j]-1;
	for (kk=0, jj=pos; kk<kk_end; jj++, kk++)
	{
		//if (s[j][kk]==txt[jj])
		if (match_table[index_table[(int)s[j][kk]]][index_table[txt[jj]]]==1)
			//if (s[j][kk+1]==txt[jj+1])
			if (match_table[index_table[(int)s[j][kk+1]]][index_table[txt[jj+1]]]==1)
			{
				double p_local = (double)di_frequencies[index_table[txt[jj]]][index_table[txt[jj+1]]]/seq_N;
				p *= p_local;
			} else {
				double p_local = (double)frequencies[index_table[txt[jj+1]]]/seq_N;					
				p *= p_local;
			}
	}
	if (p<1.)
		return ( 1 - exp( log(1-p)*L ) ); 
	else
		return 1.;
	
}

// ==============================================================
void pattern_matching::get_freq()
// ==============================================================
{
	int a,i,j;
	di_frequencies = new int*[32];
	frequencies = new int[32];

	int* frq = frequencies;
	int** di_frq = di_frequencies;
	const char* txt = T;
		
	for (i=0; i<32; i++)
	{
		/* we initialize all absolute frequencies with 4 and all
			dinucl.-freq with 1, so that we do not encounter 0
			probabilities, which may yield unwanted 0 probabilities for
			p-values. This corresponds to pretending the presence of a
			CCGGTTAACAGCTGATC in the genome sequence) */
		frq[i] = 4;
		di_frq[i] = new int[32];
		for (j=0; j<32; j++)
			di_frq[i][j] = 1;
	}
	
	int nn = N;
	for (i=0; i<nn-1; i++)
	{
		if ( (txt[i]=='C' || txt[i]=='G' || txt[i]=='T' || txt[i]=='A'
				  || txt[i]=='c' || txt[i]=='g' || txt[i]=='t' || txt[i]=='a') && 
			(txt[i+1]=='C' || txt[i+1]=='G' || txt[i+1]=='T' || txt[i+1]=='A'
				|| txt[i+1]=='c' || txt[i+1]=='g' || txt[i+1]=='t' || txt[i+1]=='a') )
		{
			di_frq[index_table[txt[i]]][index_table[txt[i+1]]]++;
			frq[index_table[txt[i]]]++;
			}
	}

	sum_up_frq('M','A','C');
	sum_up_frq('R','A','G');
	sum_up_frq('W','A','T');
	sum_up_frq('S','C','G');
	sum_up_frq('Y','C','T');
	sum_up_frq('K','G','T');
	sum_up_frq('V','M','G');
	sum_up_frq('H','M','T');
	sum_up_frq('D','R','T');
	sum_up_frq('B','S','T');
	sum_up_frq('N','M','K');
}

// ==============================================================
void pattern_matching::sum_up_frq(char O, char P, char Q)
// ==============================================================	
{
	const char* X = "ACGTMRWSYKVHDBN";
	int* frq = frequencies;
	int** di_frq = di_frequencies;
	int i;
	
	frq[index_table[O]] = frq[index_table[P]] + frq[index_table[Q]];

	for (i=0; i<15; i++)
	{
		di_frq[index_table[X[i]]][index_table[O]] =
			  di_frq[index_table[X[i]]][index_table[P]] 
			+ di_frq[index_table[X[i]]][index_table[Q]];
		
		di_frq[index_table[O]][index_table[X[i]]] =
			  di_frq[index_table[P]][index_table[X[i]]] 
			+ di_frq[index_table[Q]][index_table[X[i]]];
	}
}

// ==============================================================
void pattern_matching::compute_index_table()
// ==============================================================	
{

	int i;

	
	/// by default, we want to use asymmetric match tables, i.e.,
	/// wildcard Ns are only wildcards if they occur in a pattern
	/// fragment, but not in the text.
	bool asymmetric = true;

	index_table = new int[256];

	for(i=0; i<256; i++) index_table[i] = 30;
	
	index_table[65] = 0;   /* A */
	index_table[67] = 1;   /* C */
	index_table[71] = 2;   /* G */
	index_table[84] = 3;   /* T */
	index_table[85] = 3;   /* U == T */
	index_table[82] = 4;   /* R */
	index_table[89] = 5;   /* Y */
	index_table[77] = 6;   /* M */
	index_table[75] = 7;   /* K */
	index_table[83] = 8;   /* S */
	index_table[87] = 9;   /* W */
	index_table[72] = 10;  /* H */
	index_table[66] = 11;  /* B */
	index_table[86] = 12;  /* V */
	index_table[68] = 13;  /* D */
	if (asymmetric)
		index_table[78] = 14;  /* N */
	
	index_table[97] = 0;  /* a */
	index_table[99] = 1;  /* c */
	index_table[103] = 2; /* g  */
	index_table[116] = 3; /* t */
	index_table[117] = 3; /* u == t */
	index_table[114] = 4; /* r */
	index_table[121] = 5; /* y */
	index_table[109] = 6; /* m */
	index_table[107] = 7; /* k */
	index_table[115] = 8; /* s */
	index_table[119] = 9; /* w */
	index_table[104] = 10; /* h */
	index_table[98] = 11;  /* b */
	index_table[118] = 12; /* v */
	index_table[100] = 13; /* d */
	if (asymmetric)
		index_table[110] = 14; /* n */
	//index_table[97] = 15;  /* a */
	//index_table[99] = 16;  /* c */
	//index_table[103] = 17; /* g  */
	//index_table[116] = 18; /* t */
	//index_table[117] = 18; /* u == t */
	//index_table[114] = 19; /* r */
	//index_table[121] = 20; /* y */
	//index_table[109] = 21; /* m */
	//index_table[107] = 22; /* k */
	//index_table[115] = 23; /* s */
	//index_table[119] = 24; /* w */
	//index_table[104] = 25; /* h */
	//index_table[98] = 26;  /* b */
	//index_table[118] = 27; /* v */
	//index_table[100] = 28; /* d */
	//if (asymmetric)
		//index_table[110] = 29; /* n */

}

} // end namespace fragrep
